package fr.iutlille.ctp_2025;

import android.app.Application;
import android.util.Log;

import androidx.annotation.NonNull;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.TreeMap;
import java.util.Map;
import java.util.Objects;

public class ModelApplication extends Application {

    public static class ApplicationErrorException extends Exception {

        public ApplicationErrorException(String message) {
            super(message);
        }
    }
    private Map<String, Double> postes;
    private Map<String, Participant> participants;
    private ArrayList<Operation> operations;
    private ArrayList<String> participantsList;

    @Override
    public void onCreate() {
        super.onCreate();
        try {
            initModel();
        } catch (JSONException e) {
            Log.e("CTP", "Erreur de lecture du fichier d'initialisation...");
//            e.printStackTrace();
        }
    }

    /**
     * GESTION DES ERREURS
     */
    private void error(String message) throws ApplicationErrorException {
        Log.e ("CTP", message);
        throw new ApplicationErrorException(message);
    }

    /**
     * GESTION DES PARTICIPANTS
     */

    /**
     * Ajoute une opération
     * @param participant Le nom du participant qui fait la dépense
     * @param poste Le poste affecté
     * @param amount Le montant en euros
     * @throws ApplicationErrorException Levée si une erreur est détectée:
     * <ul>
     *     <li>Participant inexistant</li>
     *     <li>Poste inexistant</li>
     * </ul>
     */

    /**
     * Ajoute un participant au modèle.
     * Le participant est défini par son nom
     * @param name Le nom du participant ajouté
     * @throws ApplicationErrorException Détecte la tentative d'ajout alors qu'unparticipant de même nom existe déjà
     */
    public void addParticipant(String name) throws ApplicationErrorException {
        if (participants.containsKey(name)) {
            error(this.getString(R.string.ErrorInsertParticipantExists, name));
        } else {
            participants.put(name, new Participant(name));
            participantsList = new ArrayList<>(participants.keySet());
        }
    }

    /**
     * Obtient l'objet Participant à partir de son rang dans la liste (alphabétique) des paticipants..
     * @param position le rang demandé
     * @return L'objet Participant demandé
     */
    @NonNull
    public Participant getParticipant(int position) {
        return Objects.requireNonNull(participants.get(participantsList.get(position)));
    }

    /**
     * Donne le nombre de participants
      * @return Le nombre de participants
     */
    public int getNbParticipants() {
        return participantsList.size();
    }

    /**
     * Supprime un participant désigné par son nom.
     * @param name Le nom du participant à supprimer
     * @throws ApplicationErrorException. Erreurs détectées:
     * <ul>
     *     <li>Le participant n'existe pas</li>
     *     <li>Le participant a un ou des paiements à son crédit</li>
     * </ul>
     */
    public void removeParticipant(String name) throws ApplicationErrorException {
        if (! participants.containsKey(name)) {
            error(this.getString(R.string.ErrorDropUnknownParticipant, name));
        } else if (operationExists(name)) {
            error(this.getString(R.string.ErrorDropParticipantLinked));
        } else {
            participants.remove(name);
            participantsList.remove(name);
        }
    }

    /**
     * ajoute une opération.
     * @param participant le nom du participant qui a réglé la dépense.
     * @param poste le poste de la dépense
     * @param amount le montant de la dépense
     * @throws ApplicationErrorException. Erreurs détectées:
     * <ul>
     *     <li>Le participant n'existe pas</li>
     *     <li>Le poste n'existe pas</li>
     * </ul>
     */
    public void addOperation(String participant, String poste, double amount) throws ApplicationErrorException {
        if (! participants.containsKey(participant)) {
            error(this.getString(R.string.ErrorInsertUnknownParticipant, participant));
        } else if (! postes.containsKey(poste)) {
            error(this.getString(R.string.ErrorInsertUnknownPoste, poste));
        } else{
            Operation operation = new Operation(participant, poste, amount);
            operations.add(operation);
        }
    }

    /**
     * Retourne une opération définie par sa position.
     * @param position la position de l'opération demandée  (entre 0 et le nombre d'opérations-1)
     * @return l'opération recherchée.
     * @throws ApplicationErrorException Erreurs détectées:
     * <ul>
     *     <li>La position fournie est négative ou &ge; nombre d'opérations-1</li>
     * </ul>
     */
    @NonNull
    public Operation getOperation(int position) throws ApplicationErrorException {
        if ((position < 0) || (position >= operations.size())) {
            error(this.getString(R.string.ErrorOperationIllegalPosition, position));
        }
        return operations.get(position);
    }

    /**
     * Efface une opération définie par sa position.
     * @param position la position de l'opération à supprimer  (entre 0 et le nombre d'opérations-1)
     * @throws ApplicationErrorException Erreurs détectées:
     * <ul>
     *     <li>La position fournie est négative ou &ge; nombre d'opérations-1</li>
     * </ul>
     */
    void removeOperation(int position) throws ApplicationErrorException {
        if ((position < 0) || (position >= operations.size())) {
            error(this.getString(R.string.ErrorOperationIllegalPosition, position));
        } else {
            operations.remove(position);
        }
    }

    /**
     * Calcule et retourne le nombre d'opérations
     * @return le nombre d'opérations enregistrées.
     */
    public int getNbOperations() {
        return operations.size();
    }

    public double getSumOperations() {
        double sum = 0.00;
        for (Operation operation: operations) {
            sum += operation.getAmount();
        }
        return sum;
    }

    public int getNbOperations(String poste) {
        int nbOp = 0;
        for (Operation operation: operations) {
            if (operation.getPoste().equals(poste)) {
                nbOp++;
            }
        }
        return nbOp;
    }

    public double getSumOperations(String poste) {
        double sum = 0.00;
        for (Operation operation: operations) {
            if (operation.getPoste().equals(poste)) {
                sum += operation.getAmount();
            }
        }
        return sum;
    }

    /**
     * GESTION DES POSTES
     * La liste des postes est définie dans les ressources:
     * <ul>
     *     <li><code>R.string.TRANSPORT</code></li>
     *     <li><code>R.string.HEBERGEMENT</code></li>
     *     <li><code>R.string.RESTAURATION</code></li>
     *     <li><code>R.string.LOISIRS</code></li>
     *     <li><code>R.string.AUTRES</code></li>
     * </ul>
     * et dans <code>R.array.postes</code>
     */

    /**
     * Lire le cumul des dépenses pour un poste donné par son nom.
     * @param poste
     * @return
     */
    public double getSum(String poste) {
        return Objects.requireNonNullElse(this.postes.get(poste), 0.0);
    }


    /**
     *  GESTION DES INITIALISATIONS
     */
    private void initPostes() {
        String[] values = this.getResources().getStringArray(R.array.postes);
        this.postes =new TreeMap<>();
        for (String poste: values) {
            this.postes.put(poste, 0.0);
        }
    }
    private StringBuilder getConfig() {
        StringBuilder text = new StringBuilder();
        BufferedReader br = new BufferedReader(new InputStreamReader(this.getResources().openRawResource(R.raw.initial)));
        try {
            String line;
            while ((line = br.readLine()) != null) {
                text.append(line);
                text.append('\n');
            }
            br.close();
            return text;
        } catch (IOException e) {;
            e.printStackTrace();
        }
        return text;
    }

    private void initParticipants(JSONArray jsonParticipants) throws JSONException, ApplicationErrorException {
        participants = new TreeMap<>();
        for (int i = 0; i < jsonParticipants.length(); i++) {
            String name = jsonParticipants.getString(i);
            this.addParticipant(name);
        }
        participantsList = new ArrayList<>(participants.keySet());
    }

    private void initOperations(JSONArray jsonOperations) throws JSONException, ApplicationErrorException {
        operations = new ArrayList<>();
        for (int i = 0; i < jsonOperations.length(); i++) {
            JSONObject jsonOperation = jsonOperations.getJSONObject(i);
            String participant = jsonOperation.getString("participant");
            String poste = jsonOperation.getString("poste");
            double amount = jsonOperation.getDouble("amount");
            addOperation(participant, poste, amount);
        }
    }
    private void initModel() throws JSONException {
        initPostes();

        StringBuilder text = getConfig() ;
        JSONObject json = new JSONObject(text.toString());
        try {
            initParticipants(json.getJSONArray("participants"));
            initOperations(json.getJSONArray("operations"));
        } catch (ApplicationErrorException e) {
            throw new RuntimeException(e);
        }
    }

    private boolean operationExists(String name) {
        for (Operation operation: operations) {
            if (operation.getParticipant().equals(name)) return true;
        }
        return false;
    }
    /**
     * Effectue tout les calculs à partir des participants et opérations enregistrées.
     * Met à jour les objets Participant
     */
    public void computeBalance() {
        Participant.setNbParts(participants.size());
        for (Participant participant: participants.values()) {
            participant.clearBalance();
        }
        for (String poste: this.postes.keySet()) {
            postes.put(poste, 0.0);
        }
        for (Operation operation: operations) {
            Participant participant = participants.get(operation.getParticipant());
            String poste = operation.getPoste();
            double amount = Objects.requireNonNullElse(operation.getAmount(), 0.00);
            participant.addExpense(amount);
            postes.put(poste, amount + postes.get(poste));
        }
    }
}
